
#!/usr/bin/env python
# coding: utf-8

import traceback
import datetime
from BankImplementation.PricerLibrary import BlackScholes, CoxRoxRubinstein
from VectalisCore.DerivativeData import Parameter, Quote, UnderlyingReference
from VectalisCore.Error import Error

'''
To be implemented with bank specific code
'''
def evaluateQuotes(derivativeData):
    for underlying in derivativeData.underlyings:
        try:
            __retrieveMarketData(underlying)
        except Error as err:#custom error
            underlying.error = err
            print (err)
        except Exception as ex:#print the stack
            underlying.error = ex
            print (traceback.format_exc())
    
    for quote in derivativeData.quotes:
        try:
            __evaluateQuote(quote)
        except Error as err:#custom error
            quote.error = err
            print (err)
        except Exception as ex:#print the stack
            quote.error = ex
            print (traceback.format_exc())

def __retrieveMarketData(underlying):

    #get volatility surface instead of default vol
    underlying.dataReferences["Volatility"] = 20.0

    #get spot from market data source
    #underlying.dataReferences["ReferenceSpot"] = 

    #Get the Forex for instance
    #Get the Div for instance
   
def __evaluateQuote(quote):

    __setQuoteParameters(quote)

    for legConstituent in quote.legConstituents:
        leg = legConstituent.leg
        try:
            if not leg.hasError():
                __priceLeg(quote, leg)
        except Error as err:
            leg.error = err
            print (err)
        except Exception as ex:
            leg.error = ex
            print (traceback.format_exc())
    
    if not quote.hasError():
        __sumLegGreeks(quote)
        __calculateImpliedVolatility(quote)
        __setReferenceItems(quote)

def __setQuoteParameters(quote):
    if quote.singleUnderlyingReference is None: #Nothing for exotic/multi-ref quote. List of UL are stored in legConstituent.leg.underlyingDefinition
        quote.error = "Multi-underlying are not supported."
    else:
        if quote.singleUnderlyingReference.providedReferenceSpot:
            refSpot = float(quote.singleUnderlyingReference.providedReferenceSpot)
            quote.dataReferences["ReferenceSpot"] = refSpot
        elif "ReferenceSpot" in quote.singleUnderlyingReference.underlying.dataReferences:
            refSpot = float(quote.singleUnderlyingReference.underlying.dataReferences["ReferenceSpot"])
            quote.dataReferences["ReferenceSpot"] = refSpot
    
        basis = 0
        if quote.singleUnderlyingReference.underlying.typeCode == "I" and quote.singleUnderlyingReference.providedBasis:
            basis = float(quote.singleUnderlyingReference.providedBasis)
            quote.dataReferences["CashSpot"] = refSpot - basis

def __priceLeg(quote, leg):

    legDerivativeTypeId = leg.basketType + "-" + leg.derivativeTypeCode
    if legDerivativeTypeId == "SINGLE-EC":
        __priceLeg_European(quote, leg, "C", False)
    elif legDerivativeTypeId == "SINGLE-EP":
        __priceLeg_European(quote, leg, "P", False)
    elif legDerivativeTypeId == "SINGLE-EC%":
        __priceLeg_European(quote, leg, "C", True)
    elif legDerivativeTypeId == "SINGLE-EP%":
        __priceLeg_European(quote, leg, "P", True)
    elif legDerivativeTypeId == "SINGLE-AC":
        __priceLeg_American(quote, leg, "C", False)
    elif legDerivativeTypeId == "SINGLE-AP":
        __priceLeg_American(quote, leg, "P", False)
    elif legDerivativeTypeId == "SINGLE-AC%":
        __priceLeg_American(quote, leg, "C", True)
    elif legDerivativeTypeId == "SINGLE-AP%":
        __priceLeg_American(quote, leg, "P", True, )
    else :#SINGLE-_FWD%, SINGLE-_FWD, SINGLE-VARSWP, SINGLE-VOLSWP, WORSTOF-WOC%, etc
        raise Error("DerivativeType not supported: " + legDerivativeTypeId)
def __priceLeg_European(quote, leg, dt, inPercent):
    if "ReferenceSpot" not in quote.dataReferences:
        raise Error("Missing ReferenceSpot")

    today=datetime.date.today()
    expi=datetime.datetime.strptime(leg.legItems["ExpiryDate"],"%Y-%m-%d").date()
    deltaDays=expi - today
    strike=float(leg.legItems["Strike"])
    spot=float(quote.dataReferences["ReferenceSpot"])
    volatility = float(quote.singleUnderlyingReference.underlying.dataReferences["Volatility"]) #get vol from volatilifty surface

    if inPercent:
        strike = strike * spot / 100

    pricer = BlackScholes(dt, spot, strike, deltaDays.days/365,0,volatility/100)
    fairValue = pricer.fairValue()
    delta = pricer.delta()
    vega = pricer.vega()
    gamma = pricer.gamma()
    theta = pricer.theta()

    leg.returnedGreeks["FairValue"] =  Parameter(fairValue, header="FV", displayFormat="N2", backColor="#EFEFEF", participatesToSumProduct=True)
    leg.returnedGreeks["Delta"] =  Parameter(delta, header = "Delta", displayFormat="#0.%", participatesToSumProduct=True)
    leg.returnedGreeks["Gamma"] =  Parameter(gamma, header="Gamma", displayFormat="N2", participatesToSumProduct=True)
    leg.returnedGreeks["Vega"] =  Parameter(vega, header="Vega", displayFormat="N2", participatesToSumProduct=True)
    leg.returnedGreeks["Theta"] =  Parameter(theta, header="Theta", displayFormat="N2", participatesToSumProduct=True)
    leg.returnedGreeks["Volatility"] = Parameter(volatility, header="Volatility", displayFormat="N1")
def __priceLeg_American(quote, leg, dt, inPercent):
    if "ReferenceSpot" not in quote.dataReferences:
        raise Error("Missing ReferenceSpot")

    today=datetime.date.today()
    expi=datetime.datetime.strptime(leg.legItems["ExpiryDate"],"%Y-%m-%d").date()
    deltaDays=expi - today
    strike=float(leg.legItems["Strike"])
    spot=float(quote.dataReferences["ReferenceSpot"])
    volatility = float(quote.singleUnderlyingReference.underlying.dataReferences["Volatility"]) #get vol from volatilifty surface

    if inPercent:
        strike = strike * spot / 100

    pricer = CoxRoxRubinstein(dt, spot, strike, deltaDays.days/365,0,volatility/100)
    fairValue = pricer.fairValue()
    delta = pricer.delta()
    vega = pricer.vega()
    gamma = pricer.gamma()
    theta = pricer.theta()

    leg.returnedGreeks["FairValue"] =  Parameter(fairValue, header="FV", displayFormat="N2", backColor="#EFEFEF", participatesToSumProduct=True)
    leg.returnedGreeks["Delta"] =  Parameter(delta, header = "Delta", displayFormat="#0.%", participatesToSumProduct=True)
    leg.returnedGreeks["Gamma"] =  Parameter(gamma, header="Gamma", displayFormat="N2", participatesToSumProduct=True)
    leg.returnedGreeks["Vega"] =  Parameter(vega, header="Vega", displayFormat="N2", participatesToSumProduct=True)
    leg.returnedGreeks["Theta"] =  Parameter(theta, header="Theta", displayFormat="N2", participatesToSumProduct=True)
    leg.returnedGreeks["Volatility"] = Parameter(volatility, header="Volatility", displayFormat="N1")

'''
Sum leg.returnedGreeks where 'participatesToSumproduct' is true and store it as a quote.returnedGreeks
'''
def __sumLegGreeks(quote):
    leg1 = quote.legConstituents[0].leg
    for greekName, leg1Greek in leg1.returnedGreeks.items():
        if not leg1Greek.participatesToSumproduct:
            continue
        sumValue = 0
        for legConstituent in quote.legConstituents:
            greek = legConstituent.leg.returnedGreeks.get(greekName)
            if greek is None:
                sumValue = None
                break
            sumValue += (float(greek.value) * float(legConstituent.multiplier))
        if sumValue is not None:
            quote.returnedGreeks[greekName] = Parameter(sumValue, leg1Greek.header, leg1Greek.backColor, leg1Greek.displayFormat, leg1Greek.participatesToSumproduct)

def __calculateImpliedVolatility(quote):
    if quote.singleUnderlyingReference is None: #Nothing for exotic/multi-ref quote. List of UL are stored in legConstituent.leg.underlyingDefinition
        return
    
    Strike = float(quote.legConstituents[0].leg.legItems["Strike"])
    Expi = str(quote.legConstituents[0].leg.legItems["ExpiryDate"])
    HasSameExpiStrike = True
    for legConstituent in quote.legConstituents:
        if float(legConstituent.leg.legItems["Strike"]) != Strike or str(legConstituent.leg.legItems["ExpiryDate"]) != Expi:
            HasSameExpiStrike = False

    if len(quote.legConstituents) > 0 and HasSameExpiStrike == True :
        quoteVega = float(quote.legConstituents[0].leg.returnedGreeks["Vega"].value) * float(quote.legConstituents[0].multiplier)
        quoteVolatility = float(quote.legConstituents[0].leg.returnedGreeks["Volatility"].value)
        quoteFairValue = float(quote.returnedGreeks["FairValue"].value)
        for tradingInterest in quote.tradingInterests:
            impliedVolatility = quoteVolatility + (float(tradingInterest.price) - quoteFairValue) / quoteVega
            tradingInterest.returnedGreeks["ImpliedVolatility"] =  Parameter(round(impliedVolatility, 1), header="ImpliedVolatility")

'''
Add information to be read-only display on the right of the pricing datagrid
'''
def __setReferenceItems(quote):
    for (key, value) in quote.dataReferences.items():
        quote.returnedReferenceItems[key] = value
