


from math import log, sqrt, pi, exp, erf
from VectalisCore.Error import Error


class BlackScholes:
    def __init__(self, callPut, spot, strike, maturityYear, rate, vol):
        self.callPut = callPut
        self.spot = spot
        self.strike = strike
        self.maturityYear = maturityYear
        self.rate = rate
        self.vol = vol
        if self.callPut != "C" and self.callPut != "P":
            raise Error("DerivativeType not supported in BlackScholes: " + self.callPut)
    
    def fairValue(self):
        return BlackScholes.__fairValue(self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol)
    def delta(self):
        return BlackScholes.__delta(self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol)
    def vega(self):
        return BlackScholes.__fairValue(self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol + 0.01) \
                - BlackScholes.__fairValue(self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol)
    def gamma(self):
        return BlackScholes.__delta(self.callPut, self.spot*1.01, self.strike, self.maturityYear, self.rate, self.vol) \
                - BlackScholes.__delta(self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol)
    def theta(self):
        return BlackScholes.__fairValue(self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol) \
                - BlackScholes.__fairValue(self.callPut, self.spot, self.strike, self.maturityYear + 1/252, self.rate, self.vol)

    def __fairValue(callPut, spot, strike, maturityYear, rate, vol):
        if callPut == "C":
            return BlackScholes.__bs_call(spot, strike, maturityYear, rate, vol)
        elif callPut == "P":
            return BlackScholes.__bs_put(spot, strike, maturityYear, rate, vol)
    def __delta(callPut, spot, strike, maturityYear, rate, vol):
        if callPut == "C":
            return BlackScholes.__norm(BlackScholes.__d1(spot, strike, maturityYear, rate, vol))
        elif callPut == "P":
            return BlackScholes.__norm(BlackScholes.__d1(spot, strike, maturityYear, rate, vol)) - 1
    '''
    S = spot price of an asset
    K = strike price
    T = time to maturity
    r = risk-free interest rate
    σ = volatility of the asset
    '''
    def __d1(S,K,T,r,sigma):
        return(log(S/K)+(r+(sigma**2)/2.)*T)/(sigma*sqrt(T))
    def __d2(S,K,T,r,sigma):
        return BlackScholes.__d1(S,K,T,r,sigma)-sigma*sqrt(T)
    def __norm(x):
        #'Cumulative distribution function for the standard normal distribution'
        return (1.0 + erf(x / sqrt(2.0))) / 2.0

    def __bs_call(S,K,T,r,sigma):
        return S*BlackScholes.__norm(BlackScholes.__d1(S,K,T,r,sigma))-K*exp(-r*T)*BlackScholes.__norm(BlackScholes.__d2(S,K,T,r,sigma))
    def __bs_put(S,K,T,r,sigma):
        return K*exp(-r*T)-S + BlackScholes.__bs_call(S,K,T,r,sigma)



class CoxRoxRubinstein:
    def __init__(self, callPut, spot, strike, maturityYear, rate, vol):
        self.nbStep = 20
        self.callPut = callPut
        self.spot = spot
        self.strike = strike
        self.maturityYear = maturityYear
        self.rate = rate
        self.vol = vol
        if self.callPut != "C" and self.callPut != "P":
            raise Error("DerivativeType not supported in BlackScholes: " + self.callPut)
    
    def fairValue(self):
        return CoxRoxRubinstein.__binomial(self.nbStep, self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol)
    def delta(self):
        return (CoxRoxRubinstein.__binomial(self.nbStep, self.callPut, self.spot*1.001, self.strike, self.maturityYear, self.rate, self.vol) \
                - CoxRoxRubinstein.__binomial(self.nbStep, self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol)) / (self.spot*0.001)
    def vega(self):
        return CoxRoxRubinstein.__binomial(self.nbStep, self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol + 0.01) \
                - CoxRoxRubinstein.__binomial(self.nbStep, self.callPut, self.spot, self.strike, self.maturityYear, self.rate, self.vol)
    def gamma(self):
        return 0.02 #dummy value
    def theta(self):
        return 0.02 #dummy value

    '''
    dt = call or put
    n = number of steps
    S = spot price of an asset
    K = strike price
    T = time to maturity
    r = risk-free interest rate
    v = volatility of the asset
    '''
    def __binomial(n, dt, S, K, T, r, v):  
        At = T/n 
        u = exp(v*sqrt(At))
        d = 1./u
        p = (exp(r*At)-d) / (u-d) 
        #Binomial price tree
        stockvalue = [[0]*(n+1) for x in range(n+1)]
        stockvalue[0][0] = S
        for i in range(1,n+1):
            stockvalue[i][0] = stockvalue[i-1][0]*u
            for j in range(1,i+1):
                stockvalue[i][j] = stockvalue[i-1][j-1]*d
        #option value at final node   
        optionvalue = [[0]*(n+1) for x in range(n+1)]
        for j in range(n+1):
            if dt=="C": # Call
                optionvalue[n][j] = max(0, stockvalue[n][j]-K)
            elif dt =="P": #Put
                optionvalue[n][j] = max(0, K-stockvalue[n][j])
        #backward calculation for option price    
        for i in range(n-1,-1,-1):
            for j in range(i+1):
                    if dt=="P":
                        optionvalue[i][j] = max(0, K-stockvalue[i][j], exp(-r*At)*(p*optionvalue[i+1][j]+(1-p)*optionvalue[i+1][j+1]))
                    elif dt=="C":
                        optionvalue[i][j] = max(0, stockvalue[i][j]-K, exp(-r*At)*(p*optionvalue[i+1][j]+(1-p)*optionvalue[i+1][j+1]))
        return optionvalue[0][0]
