#!/usr/bin/env python
# coding: utf-8

from collections import OrderedDict

# Business objects:
# DerivativeData
#   |-- Underlying (a Share or an Equity Index)
#   |    ^
#   |    |
#   |-- UnderlyingReference (Underlying with a ReferenceSpot)
#   |     ^ 
#   |     |
#   |    UnderlyingConstituent (include a potential weight)
#   |     |  
#   |-- UnderlyingDefinition (a single Underlying or a basket of Underlyings)
#   |    ^
#   |    |
#   |-- Leg  (primitive derivative as a Call or a Varswap or a Call-on-BestOf)   
#   |     ^ 
#   |     |
#   |    LegConstituent (include a multitplier)
#   |     |
#   |-- Quote (a strategy)
#
class Parameter:
    def __init__(self, value, header: str = None, foreColor: str = None, backColor: str = None, displayFormat: str = None, participatesToSumProduct: bool = False):
        self.value = value
        self.header = header
        self.foreColor = foreColor
        self.backColor = backColor
        self.displayFormat = displayFormat
        self.participatesToSumproduct = participatesToSumProduct

class DerivativeData:
    def __init__(self):
        self.underlyings = []
        self.underlyingReferences = []
        self.underlyingDefinitions = []
        self.quotes = []
        self.request = None
 
class Underlying:
    def __init__(self):
        self.vectalisCode = None
        self.pricerCode = None
        self.typeCode = None#I or S
        self.currencyCode = None
        self.displayCode = None
        self.reutersCode = None
        self.bloombergCode = None

        self.dataReferences = {}
        self.error = None
    def __str__(self):
        return self.vectalisCode
    def getIdString(self):
        return self.vectalisCode
    def getError(self):
        return self.error
    def hasError(self):
        return self.getError() is not None
class UnderlyingReference:
    def __init__(self):
        self.underlying = None
        self.providedReferenceSpot = None
        self.providedFutureExpiryDate = None
        self.providedBasis = None
        self.providedDelta = None
        self.error = None
    def getIdString(self):
        return self.underlying.getIdString() + " - ref=" + str(self.providedReferenceSpot) + " " + str(self.providedFutureExpiryDate)
    def getError(self):
        if self.error is not None:
            return self.error
        uError = self.underlying.getError()
        if uError is not None:
            return uError
        return None
    def hasError(self):
        return self.getError() is not None
class UnderlyingConstituent:
    def __init__(self):
        self.underlyingReference = None
        self.weight = None
        self.error = None
    def getError(self):
        if self.error is not None:
            return self.error
        uError = self.underlyingReference.getError()
        if uError is not None:
            return uError
        return None
    def hasError(self):
        return self.getError() is not None
class UnderlyingDefinition:
    def __init__(self):
        self.name = None
        self.displayName = None
        self.hasWeight = False
        self.constituents = []
        self.error = None
    def getIdString(self):
        r = ""
        for c in self.constituents:
            r = r + c.underlyingReference.getIdString() + " w=" + str(c.weight) + " "
        return "(" + r + ")"
    def getError(self):
        if self.error is not None:
            return self.error
        for c in self.constituents:
            uError = c.underlyingReference.getError()
            if uError is not None:
                return uError
        return None
    def hasError(self):
        return self.getError() is not None
class Leg:
    def __init__(self):
        self.basketType = None
        self.derivativeTypeCode = None
        self.isPriceInPercent = False
        self.underlyingDefinition = None
        self.legItems = {}
        self.returnedGreeks = OrderedDict()
        self.error = None
    def getIdString(self):
        return self.basketType + "-" + self.derivativeTypeCode + " " + self.underlyingDefinition.getIdString() + " " + str(self.legItems)
    def getError(self):
        if self.error is not None:
            return self.error
        uError = self.underlyingDefinition.getError()
        if uError is not None:
            return uError
        return None
    def hasError(self):
        return self.getError() is not None
class LegConstituent:
    def __init__(self):
        self.leg: Leg = None
        self.index = None
        self.multiplier = None
class TradingInterest:
    def __init__(self):
        self.code = None
        self.price = None
        self.size = None
        self.time = None
        self.conditionString = None
        self.comment = None
        self.returnedGreeks = OrderedDict()

class SalesTradeInfo:
    def __init__(self):
        self.salesOtcSize = None
        self.salesListedSize = None 
        self.salesPandL = None # P&L    
        self.salesClient = None
        self.salesComment = None
        self.salesCustomString1 = None
        self.salesCustomString2 = None
        self.salesCustomString3 = None
        self.salesCustomString4 = None
        self.salesCustomString5 = None
        self.salesCustomString6 = None
        self.salesCustomDouble1 = None
        self.salesCustomDouble2 = None
        self.salesCustomBool1 = None
        self.salesCustomBool2 = None
class CustomTradeInfo:
    def _init_(self):
        self.customString1 = None
        self.customString2 = None
        self.customString3 = None
        self.customString4 = None
        self.customString5 = None
        self.customDouble1 = None
        self.customDouble2 = None
        self.customDouble3 = None
        self.customDouble4 = None
        self.customDouble5 = None
        self.customBool1 = None
        self.customBool2 = None
        self.customBool3 = None
        self.customBool4 = None
        self.customBool5 = None
class Quote:
    def __init__(self):
        self.quoteMessageId = None
        self.quoteId = None
        self.legConstituents = []
        self.tradingInterests = []
        self.derivativeTypeCode = None
        self.derivativeTypePricerCode = None
        self.derivativeTypeName = None
        self.listedOrOtc = None
        self.quotationType = None
        self.deferredPremium = None
        self.providerDisplayCode = None
        self.providerEntityCode = None
        self.providerCode = None
        self.providerEntityName = None
        self.providerName = None
        self.providerComment = None
        self.overwriteComment = None
        self.packageId = None
        self.packageMultiplier = None
        self.isPackageChoice = None
        self.salesTradeInfo = None
        self.customTradeInfo = None
        self.singleUnderlyingReference = None #Nothing for exotic/multi-ref quote. List of UL are stored in legConstituent.leg.underlyingDefinition
        self.dataReferences = {}

        self.returnedGreeks = OrderedDict()
        self.returnedReferenceItems = OrderedDict()
        self.error = None
    def getError(self):
        if self.error is not None:
            return self.error
        for l in self.legConstituents:
            lError = l.leg.getError()
            if lError is not None:
                return lError
        return None
    def hasError(self):
        return self.getError() is not None
class Request:
    def __init__(self):
        self.xmlPricingMessageVersion = None # "4.1"
        self.requesterEntityCode = None
        self.requesterCode = None #user code
        self.requesterAddresseeCode = None #work as
        self.windowsUsername = None
        self.operationDate = None
        self.pricingParameters = None #Values from UTS, set in Configuration->Pricing->PrincingParameters
class CustomCommandRequest:
    def __init__(self):
        self.command = None
        self.derivativeData = None




